/*
 * ipfilter.{cc,hh} -- IP-packet filter with tcpdumplike syntax
 * Eddie Kohler
 *
 * Copyright (c) 2000-2004 Mazu Networks, Inc.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, subject to the conditions
 * listed in the Click LICENSE file. These conditions include: you must
 * preserve this copyright notice, and you cannot mention the copyright
 * holders in advertising related to the Software without their permission.
 * The Software is provided WITHOUT ANY WARRANTY, EXPRESS OR IMPLIED. This
 * notice is a summary of the Click LICENSE file; the license in that file is
 * legally binding.
 */

#include <click/config.h>
#include "ipfilter.hh"
#include <click/glue.hh>
#include <click/error.hh>
#include <click/confparse.hh>
#include <click/straccum.hh>
#include <clicknet/ip.h>
#include <clicknet/tcp.h>
#include <clicknet/icmp.h>
#include <click/hashmap.hh>
#include <click/integers.hh>
CLICK_DECLS

static IPFilter::Wordmap* wordmap;

IPFilter::Wordmap::Wordmap()
  : _map(0)
{
}

void
IPFilter::Wordmap::insert(const String &name, uint32_t type, uint32_t data)
{
  if (type <= 0x7FFF && data <= 0xFFFF)
    _map.insert(name, (type << 16) | data);
  else {
    _map.insert(name, 0x80000000 | _type.size());
    _type.push_back(type);
    _data.push_back(data);
  }
}

IPFilter::Wordmap*
IPFilter::create_wordmap()
{
  IPFilter::Wordmap* wordmap = new Wordmap;

  wordmap->insert("host",	TYPE_TYPE, TYPE_HOST);
  wordmap->insert("net",	TYPE_TYPE, TYPE_NET);
  wordmap->insert("port",	TYPE_TYPE, TYPE_PORT);
  wordmap->insert("proto",	TYPE_TYPE, TYPE_PROTO);
  wordmap->insert("opt",	TYPE_TYPE, TYPE_TCPOPT);
  wordmap->insert("tos",	TYPE_TYPE, FIELD_TOS);
  wordmap->insert("ttl",	TYPE_TYPE, FIELD_TTL);
  wordmap->insert("dscp",	TYPE_TYPE, FIELD_DSCP);
  wordmap->insert("type",	TYPE_TYPE, FIELD_ICMP_TYPE);
  wordmap->insert("frag",	TYPE_TYPE, TYPE_IPFRAG);
  wordmap->insert("unfrag",	TYPE_TYPE, TYPE_IPUNFRAG);
  wordmap->insert("ect",	TYPE_TYPE, TYPE_IPECT);
  wordmap->insert("ce",		TYPE_TYPE, TYPE_IPCE);
  wordmap->insert("len",	TYPE_TYPE, FIELD_IPLEN);
  wordmap->insert("vers",	TYPE_TYPE, FIELD_VERSION);
  wordmap->insert("hl",		TYPE_TYPE, FIELD_HL);
  wordmap->insert("id",		TYPE_TYPE, FIELD_ID);
  wordmap->insert("win",	TYPE_TYPE, FIELD_TCP_WIN);
  
  wordmap->insert("icmp",	TYPE_PROTO, IP_PROTO_ICMP);
  wordmap->insert("igmp",	TYPE_PROTO, IP_PROTO_IGMP);
  wordmap->insert("ipip",	TYPE_PROTO, IP_PROTO_IPIP);
  wordmap->insert("tcp",	TYPE_PROTO, IP_PROTO_TCP);
  wordmap->insert("udp",	TYPE_PROTO, IP_PROTO_UDP);
  wordmap->insert("tcpudp",	TYPE_PROTO, IP_PROTO_TCP_OR_UDP);
  wordmap->insert("transp",	TYPE_PROTO, IP_PROTO_NONE);
  
  wordmap->insert("echo",	TYPE_PORT, 7);
  wordmap->insert("discard",	TYPE_PORT, 9);
  wordmap->insert("daytime",	TYPE_PORT, 13);
  wordmap->insert("chargen",	TYPE_PORT, 19);
  wordmap->insert("ftp-data",	TYPE_PORT, 20);
  wordmap->insert("ftp",	TYPE_PORT, 21);
  wordmap->insert("ssh",	TYPE_PORT, 22);
  wordmap->insert("telnet",	TYPE_PORT, 23);
  wordmap->insert("smtp",	TYPE_PORT, 25);
  wordmap->insert("domain",	TYPE_PORT, 53);
  wordmap->insert("dns",	TYPE_PORT, 53);
  wordmap->insert("bootps",	TYPE_PORT, 67);
  wordmap->insert("bootpc",	TYPE_PORT, 68);
  wordmap->insert("tftp",	TYPE_PORT, 69);
  wordmap->insert("finger",	TYPE_PORT, 79);
  wordmap->insert("www",	TYPE_PORT, 80);
  wordmap->insert("pop3",	TYPE_PORT, 110);
  wordmap->insert("sunrpc",	TYPE_PORT, 111);
  wordmap->insert("auth",	TYPE_PORT, 113);
  wordmap->insert("nntp",	TYPE_PORT, 119);
  wordmap->insert("ntp",	TYPE_PORT, 123);
  wordmap->insert("netbios-ns",	TYPE_PORT, 137);
  wordmap->insert("netbios-dgm",TYPE_PORT, 138);
  wordmap->insert("netbios-ssn",TYPE_PORT, 139);
  wordmap->insert("snmp",	TYPE_PORT, 161);
  wordmap->insert("snmp-trap",	TYPE_PORT, 162);
  wordmap->insert("irc",	TYPE_PORT, 194);
  wordmap->insert("imap3",	TYPE_PORT, 220);
  wordmap->insert("https",	TYPE_PORT, 443);
  wordmap->insert("rip",	TYPE_PORT, 520);
  wordmap->insert("route",	TYPE_PORT, 520);
  wordmap->insert("imaps",	TYPE_PORT, 993);
  wordmap->insert("pop3s",	TYPE_PORT, 995);

  wordmap->insert("syn",	TYPE_TCPOPT, TH_SYN);
  wordmap->insert("fin",	TYPE_TCPOPT, TH_FIN);
  wordmap->insert("ack",	TYPE_TCPOPT, TH_ACK);
  wordmap->insert("rst",	TYPE_TCPOPT, TH_RST);
  wordmap->insert("psh",	TYPE_TCPOPT, TH_PUSH);
  wordmap->insert("urg",	TYPE_TCPOPT, TH_URG);

  wordmap->insert("echo-reply",	FIELD_ICMP_TYPE, ICMP_ECHOREPLY);
  wordmap->insert("unreachable", FIELD_ICMP_TYPE, ICMP_UNREACH);
  wordmap->insert("sourcequench", FIELD_ICMP_TYPE, ICMP_SOURCEQUENCH);
  wordmap->insert("redirect",	FIELD_ICMP_TYPE, ICMP_REDIRECT);
  wordmap->insert("echo+",	FIELD_ICMP_TYPE, ICMP_ECHO);
  wordmap->insert("routeradvert", FIELD_ICMP_TYPE, ICMP_ROUTERADVERT);
  wordmap->insert("routersolicit", FIELD_ICMP_TYPE, ICMP_ROUTERSOLICIT);
  wordmap->insert("timeexceeded", FIELD_ICMP_TYPE, ICMP_TIMXCEED);
  wordmap->insert("parameterproblem", FIELD_ICMP_TYPE, ICMP_PARAMPROB);
  wordmap->insert("timestamp",	FIELD_ICMP_TYPE, ICMP_TSTAMP);
  wordmap->insert("timestamp-reply", FIELD_ICMP_TYPE, ICMP_TSTAMPREPLY);
  wordmap->insert("inforeq",	FIELD_ICMP_TYPE, ICMP_IREQ);
  wordmap->insert("inforeq-reply", FIELD_ICMP_TYPE, ICMP_IREQREPLY);
  wordmap->insert("maskreq",	FIELD_ICMP_TYPE, ICMP_MASKREQ);
  wordmap->insert("maskreq-reply", FIELD_ICMP_TYPE, ICMP_MASKREQREPLY);

  // deprecated variants
  wordmap->insert("echo_reply",	FIELD_ICMP_TYPE, ICMP_ECHOREPLY);
  wordmap->insert("dst_unreachable", FIELD_ICMP_TYPE, ICMP_UNREACH);
  wordmap->insert("source_quench", FIELD_ICMP_TYPE, ICMP_SOURCEQUENCH);
  wordmap->insert("time_exceeded", FIELD_ICMP_TYPE, ICMP_TIMXCEED);
  wordmap->insert("parameter_problem", FIELD_ICMP_TYPE, ICMP_PARAMPROB);
  wordmap->insert("time_stamp", FIELD_ICMP_TYPE, ICMP_TSTAMP);
  wordmap->insert("time_stamp_reply", FIELD_ICMP_TYPE, ICMP_TSTAMPREPLY);
  wordmap->insert("info_request", FIELD_ICMP_TYPE, ICMP_IREQ);
  wordmap->insert("info_request_reply", FIELD_ICMP_TYPE, ICMP_IREQREPLY);

  return wordmap;
}

void
IPFilter::Wordmap::accum_names(String word, StringAccum &sa) const
{
  int t = _map[word];
  Vector<int> types;
  while (t != 0) {
    int t_type = (t & 0x80000000 ? _type[t & 0x7FFFFFFF] : (t >> 16));
    types.push_back(t_type);
    word += "+";
    t = _map[word];
  }

  if (types.size() == 0)
    /* nada */;
  else if (types.size() == 1)
    sa << '\'' << IPFilter::Primitive::unparse_type(0, types[0]) << '\'';
  else if (types.size() == 2)
    sa << '\'' << IPFilter::Primitive::unparse_type(0, types[0]) << "' or '" << IPFilter::Primitive::unparse_type(0, types[1]) << '\'';
  else {
    for (int i = 0; i < types.size() - 1; i++)
      sa << '\'' << IPFilter::Primitive::unparse_type(0, types[i]) << "', ";
    sa << "or '" << IPFilter::Primitive::unparse_type(0, types.back()) << '\'';
  }
}

static bool
protos_compatible(int p1, int p2)
{
  return (p1 == p2
	  || (p1 == IP_PROTO_TCP_OR_UDP && (p2 == IP_PROTO_TCP || p2 == IP_PROTO_UDP))
	  || (p2 == IP_PROTO_TCP_OR_UDP && (p1 == IP_PROTO_TCP || p2 == IP_PROTO_UDP)));
}

int
IPFilter::Wordmap::lookup(String word, int type, int transp_proto, uint32_t &data, ErrorHandler *errh) const
{
  String orig_word = word;
  int t = _map[word];
  if (t == 0)
    return -1;

  int num_matches = 0;
  int last_t_match = -1;
  uint32_t last_d_match = 0;
  int transp_dropped = 0;

  while (t != 0) {
    // unpack into type and data
    int t_type, t_data;
    if (t & 0x80000000)
      t_type = _type[t & 0x7FFFFFFF], t_data = _data[t & 0x7FFFFFFF];
    else
      t_type = (t >> 16), t_data = (t & 0xFFFF);

    // what transport protocol does this type imply?
    int t_proto = 0;
    if (t_type == TYPE_TCPOPT)
      t_proto = IP_PROTO_TCP;
    else if (t_type == TYPE_PORT)
      t_proto = IP_PROTO_TCP_OR_UDP;
    else if (t_type & TYPE_FIELD)
      t_proto = (t_type & FIELD_PROTO_MASK) >> FIELD_PROTO_SHIFT;
    
    if (type != 0 && t_type != type)
      /* not interesting */;
    else if (t_proto && transp_proto != UNKNOWN && !protos_compatible(t_proto, transp_proto)) {
      /* not interesting */
      transp_dropped++;
    } else {
      num_matches++;
      last_t_match = t_type;
      last_d_match = t_data;
    }

    word += "+";
    t = _map[word];
  }

  if (num_matches == 1) {
    data = last_d_match;
    return last_t_match;
  } else if (num_matches > 1) {
    if (errh) {
      StringAccum sa;
      accum_names(orig_word, sa);
      errh->error("'%s' is ambiguous; specify %s", orig_word.cc(), sa.cc());
    }
    return -2;
  } else {
    if (errh) {
      String tn = Primitive::unparse_type(0, type);
      String tr = Primitive::unparse_transp_proto(transp_proto);
      if (tr) tr += " ";
      StringAccum sa;
      accum_names(orig_word, sa);
      errh->error("'%s%s %s' is meaningless; specify %s with %s", tr.cc(), tn.cc(), orig_word.cc(), sa.cc(), orig_word.cc());
    }
    return -2;
  }
}

void
IPFilter::static_initialize()
{
  wordmap = create_wordmap();
}

void
IPFilter::static_cleanup()
{
  delete wordmap;
  wordmap = 0;
}


IPFilter::IPFilter()
{
  // no MOD_INC_USE_COUNT; rely on Classifier
}

IPFilter::~IPFilter()
{
  // no MOD_DEC_USE_COUNT; rely on Classifier
}

//
// CONFIGURATION
//

void
IPFilter::Primitive::clear()
{
  _type = _srcdst = 0;
  _transp_proto = UNKNOWN;
  _data = 0;
  _op = OP_EQ;
  _op_negated = false;
}

void
IPFilter::Primitive::set_type(int x, ErrorHandler *errh)
{
  if (_type)
    errh->error("type specified twice");
  _type = x;
}

void
IPFilter::Primitive::set_srcdst(int x, ErrorHandler *errh)
{
  if (_srcdst)
    errh->error("'src' or 'dst' specified twice");
  _srcdst = x;
}

void
IPFilter::Primitive::set_transp_proto(int x, ErrorHandler *errh)
{
  if (_transp_proto != UNKNOWN && _transp_proto != x)
    errh->error("transport protocol specified twice");
  _transp_proto = x;
}

int
IPFilter::Primitive::set_mask(uint32_t full_mask, int shift, ErrorHandler *errh)
{
  uint32_t data = _u.u;

  // Two kinds of GT or LT tests are OK: those that boil down to checking
  // the upper bits against 0, and those that boil down to checking the upper
  // bits against ~0.
  if (_op == OP_GT || _op == OP_LT) {
    // Check for comparisons that are always true or false.
    if ((_op == OP_LT && (data == 0 || data > full_mask))
	|| (_op == OP_GT && data >= full_mask)) {
      bool will_be = (_op == OP_LT && data > full_mask ? !_op_negated : _op_negated);
      errh->warning("relation '%s %u' is always %s (range 0-%u)", unparse_op().cc(), data, (will_be ? "true" : "false"), full_mask);
      _u.u = _mask.u = 0;
      _op_negated = !will_be;
      _op = OP_EQ;
      return 0;
    }

    // value < X == !(value > (X - 1))
    if (_op == OP_LT) {
      _u.u--;
      _op_negated = !_op_negated;
      _op = OP_GT;
    }

    _u.u = (_u.u << shift) | ((1 << shift) - 1);
    _mask.u = (full_mask << shift) | ((1 << shift) - 1);
    return 0;
  }

  if (data > full_mask)
    return errh->error("value %u out of range (0-%u)", data, full_mask);

  _u.u = data << shift;
  _mask.u = full_mask << shift;
  return 0;
}

String
IPFilter::Primitive::unparse_type(int srcdst, int type)
{
  StringAccum sa;
  
  switch (srcdst) {
   case SD_SRC: sa << "src "; break;
   case SD_DST: sa << "dst "; break;
   case SD_OR: sa << "src or dst "; break;
   case SD_AND: sa << "src and dst "; break;
  }
  
  switch (type) {
   case TYPE_NONE: sa << "<none>"; break;
   case TYPE_HOST: sa << "ip host"; break;
   case TYPE_PROTO: sa << "proto"; break;
   case TYPE_IPFRAG: sa << "ip frag"; break;
   case TYPE_PORT: sa << "port"; break;
   case TYPE_TCPOPT: sa << "tcp opt"; break;
   case TYPE_NET: sa << "ip net"; break;
   case TYPE_IPUNFRAG: sa << "ip unfrag"; break;
   case TYPE_IPECT: sa << "ip ect"; break;
   case TYPE_IPCE: sa << "ip ce"; break;
   default:
    if (type & TYPE_FIELD) {
      switch (type) {
       case FIELD_IPLEN: sa << "ip len"; break;
       case FIELD_ID: sa << "ip id"; break;
       case FIELD_VERSION: sa << "ip vers"; break;
       case FIELD_HL: sa << "ip hl"; break;
       case FIELD_TOS: sa << "ip tos"; break;
       case FIELD_DSCP: sa << "ip dscp"; break;
       case FIELD_TTL: sa << "ip ttl"; break;
       case FIELD_TCP_WIN: sa << "tcp win"; break;
       case FIELD_ICMP_TYPE: sa << "icmp type"; break;
       default:
	if (type & FIELD_PROTO_MASK)
	  sa << unparse_transp_proto((type & FIELD_PROTO_MASK) >> FIELD_PROTO_SHIFT);
	else
	  sa << "ip";
	sa << "[...]";
	break;
      }
    } else
      sa << "<unknown type " << type << ">";
    break;
  }

  return sa.take_string();
}

String
IPFilter::Primitive::unparse_transp_proto(int transp_proto)
{
  switch (transp_proto) {
   case UNKNOWN: return "";
   case IP_PROTO_ICMP: return "icmp";
   case IP_PROTO_IGMP: return "igmp";
   case IP_PROTO_IPIP: return "ipip";
   case IP_PROTO_TCP: return "tcp";
   case IP_PROTO_UDP: return "udp";
   case IP_PROTO_TCP_OR_UDP: return "tcpudp";
   case IP_PROTO_NONE: return "transp";
   default: return "ip proto " + String(transp_proto);
  }
}

String
IPFilter::Primitive::unparse_type() const
{
  return unparse_type(_srcdst, _type);
}

String
IPFilter::Primitive::unparse_op() const
{
  if (_op == OP_GT)
    return (_op_negated ? "<=" : ">");
  else if (_op == OP_LT)
    return (_op_negated ? ">=" : "<");
  else
    return (_op_negated ? "!=" : "=");
}

void
IPFilter::Primitive::simple_negate()
{
  assert(negation_is_simple());
  _op_negated = !_op_negated;
  if (_type == TYPE_PROTO && _mask.u == 0xFF)
    _transp_proto = (_op_negated ? UNKNOWN : _u.i);
}

int
IPFilter::Primitive::check(const Primitive &p, ErrorHandler *errh)
{
  int old_srcdst = _srcdst;

  // if _type is erroneous, return -1 right away
  if (_type < 0)
    return -1;
  
  // set _type if it was not specified
  if (!_type) {

   retry:
    switch (_data) {

     case TYPE_HOST:
     case TYPE_NET:
     case TYPE_TCPOPT:
      _type = _data;
      if (!_srcdst)
	_srcdst = p._srcdst;
      break;

     case TYPE_PROTO:
      _type = TYPE_PROTO;
      break;
      
     case TYPE_PORT:
      _type = TYPE_PORT;
      if (!_srcdst)
	_srcdst = p._srcdst;
      if (_transp_proto == UNKNOWN)
	_transp_proto = p._transp_proto;
      break;
      
     case TYPE_INT:
      if (!(p._type & TYPE_FIELD) && p._type != TYPE_PORT)
	return errh->error("specify header field or 'port'");
      _data = p._type;
      goto retry;

     case TYPE_NONE:
      if (_transp_proto != UNKNOWN)
	_type = TYPE_PROTO;
      else
	return errh->error("partial directive");
      break;
      
     default:
      if (_data & TYPE_FIELD) {
	_type = _data;
	if ((_type & FIELD_PROTO_MASK) && _transp_proto == UNKNOWN)
	  _transp_proto = (_type & FIELD_PROTO_MASK) >> FIELD_PROTO_SHIFT;
      } else
	return errh->error("unknown type '%s'", unparse_type(0, _data).cc());
      break;

    }
  }

  // check that _data and _type agree
  switch (_type) {

   case TYPE_HOST:
    if (_data != TYPE_HOST)
      return errh->error("IP address missing in 'host' directive");
    if (_op != OP_EQ)
      return errh->error("can't use relational operators with 'host'");
    _mask.u = 0xFFFFFFFFU;
    break;

   case TYPE_NET:
    if (_data != TYPE_NET)
      return errh->error("IP prefix missing in 'net' directive");
    if (_op != OP_EQ)
      return errh->error("can't use relational operators with 'net'");
    _type = TYPE_HOST;
    // _mask already set
    break;

   case TYPE_PROTO:
    if (_data == TYPE_INT || _data == TYPE_PROTO) {
      if (_transp_proto != UNKNOWN && _transp_proto != _u.i)
	return errh->error("transport protocol specified twice");
      _data = TYPE_NONE;
    } else
      _u.i = _transp_proto;
    _transp_proto = UNKNOWN;
    if (_data != TYPE_NONE || _u.i == UNKNOWN)
      return errh->error("IP protocol missing in 'proto' directive");
    if (_u.i >= 256) {
      if (_op != OP_EQ)
	return errh->error("can't use relational operators with '%s'", unparse_transp_proto(_u.i).c_str());
      _mask.u = 0xFF;
    } else if (set_mask(0xFF, 0, errh) < 0)
      return -1;
    if (_op == OP_EQ && _mask.u == 0xFF && !_op_negated) // set _transp_proto if allowed
      _transp_proto = _u.i;
    break;

   case TYPE_PORT:
    if (_data == TYPE_INT)
      _data = TYPE_PORT;
    if (_data != TYPE_PORT)
      return errh->error("port number missing in 'port' directive");
    if (_transp_proto == UNKNOWN)
      _transp_proto = IP_PROTO_TCP_OR_UDP;
    else if (_transp_proto != IP_PROTO_TCP && _transp_proto != IP_PROTO_UDP && _transp_proto != IP_PROTO_TCP_OR_UDP)
      return errh->error("bad protocol %d for 'port' directive", _transp_proto);
    if (set_mask(0xFFFF, 0, errh) < 0)
      return -1;
    break;

   case TYPE_TCPOPT:
    if (_data == TYPE_INT)
      _data = TYPE_TCPOPT;
    if (_data != TYPE_TCPOPT)
      return errh->error("TCP options missing in 'tcp opt' directive");
    if (_transp_proto == UNKNOWN)
      _transp_proto = IP_PROTO_TCP;
    else if (_transp_proto != IP_PROTO_TCP)
      return errh->error("bad protocol %d for 'tcp opt' directive", _transp_proto);
    if (_op != OP_EQ || _op_negated)
      return errh->error("can't use relational operators with 'tcp opt'");
    if (_u.i < 0 || _u.i > 255)
      return errh->error("value %d out of range", _u.i);
    _mask.i = _u.i;
    break;

   case TYPE_IPECT:
    if (_data == TYPE_NONE) {
      _mask.u = IP_ECNMASK;
      _u.u = 0;
      _op_negated = true;
    } else if (_data == TYPE_INT) {
      if (set_mask(0x3, 0, errh) < 0)
	return -1;
    } else
      return errh->error("weird data given to 'ip ect' directive");
    _type = FIELD_TOS;
    break;

   case TYPE_IPCE:
    if (_data != TYPE_NONE)
      return errh->error("'ip ce' directive takes no data");
    _mask.u = IP_ECNMASK;
    _u.u = IP_ECN_CE;
    _type = FIELD_TOS;
    break;

   case TYPE_IPFRAG:
    if (_data != TYPE_NONE)
      return errh->error("'ip frag' directive takes no data");
    _mask.u = 1; // don't want mask to be 0
    break;

   case TYPE_IPUNFRAG:
    if (_data != TYPE_NONE)
      return errh->error("'ip unfrag' directive takes no data");
    _op_negated = true;
    _mask.u = 1; // don't want mask to be 0
    _type = TYPE_IPFRAG;
    break;

   default:
    if (_type & TYPE_FIELD) {
      if (_data != TYPE_INT && _data != _type)
	return errh->error("value missing in '%s' directive", unparse_type().c_str());
      int nbits = ((_type & FIELD_LENGTH_MASK) >> FIELD_LENGTH_SHIFT) + 1;
      uint32_t mask = (nbits == 32 ? 0xFFFFFFFFU : (1 << nbits) - 1);
      if (set_mask(mask, 0, errh) < 0)
	return -1;
    }
    break;
    
  }

  // fix _srcdst
  if (_type == TYPE_HOST || _type == TYPE_PORT) {
    if (_srcdst == 0)
      _srcdst = SD_OR;
  } else if (old_srcdst)
    errh->warning("'src' or 'dst' is meaningless here");
  
  return 0;
}

static void
add_exprs_for_proto(int32_t proto, int32_t mask, Classifier *c, Vector<int> &tree)
{
  if (mask == 0xFF && proto == IP_PROTO_TCP_OR_UDP) {
    c->start_expr_subtree(tree);
    c->add_expr(tree, 8, htonl(IP_PROTO_TCP << 16), htonl(0x00FF0000));
    c->add_expr(tree, 8, htonl(IP_PROTO_UDP << 16), htonl(0x00FF0000));
    c->finish_expr_subtree(tree, Classifier::C_OR);
  } else if (mask == 0xFF && proto == IP_PROTO_NONE)
    /* nada */;
  else
    c->add_expr(tree, 8, htonl(proto << 16), htonl(mask << 16));
}

void
IPFilter::Primitive::add_comparison_exprs(Classifier *c, Vector<int> &tree, int offset, int shift, bool swapped) const
{
  assert(_op == IPFilter::OP_EQ || _op == IPFilter::OP_GT);

  uint32_t mask = _mask.u;
  uint32_t u = _u.u & mask;
  if (swapped) {
    mask = ntohl(mask);
    u = ntohl(u);
  }

  if (_op == IPFilter::OP_EQ) {
    c->add_expr(tree, offset, htonl(u << shift), htonl(mask << shift));
    if (_op_negated)
      c->negate_expr_subtree(tree);
    return;
  }

  // To implement a greater-than test for "input&MASK > U":
  // Check the top bit of U&MASK.
  // If the top bit is 0, then:
  //    Find TOPMASK, the top bits of MASK s.t. U&TOPMASK == 0.
  //    If "input&TOPMASK == 0", continue testing with lower bits of
  //    U and MASK; combine with OR.
  //    Otherwise, succeed.
  // If the top bit is 1, then:
  //    Find TOPMASK, the top bits of MASK s.t. (U+1)&TOPMASK == TOPMASK.
  //    If "input&TOPMASK == TOPMASK", continue testing with lower bits of
  //    U and MASK; combine with AND.
  //    Otherwise, fail.
  // Stop testing when U >= MASK.
  
  int high_bit_record = 0;
  int count = 0;

  while (u < mask) {
    int high_bit = (u > (mask >> 1));
    int first_different_bit = 33 - ffs_msb(high_bit ? ~(u+1) & mask : u);
    uint32_t upper_mask;
    if (first_different_bit == 33)
      upper_mask = mask;
    else
      upper_mask = mask & ~((1 << first_different_bit) - 1);
    uint32_t upper_u = (high_bit ? 0xFFFFFFFF & upper_mask : 0);

    c->start_expr_subtree(tree);
    c->add_expr(tree, offset, htonl(upper_u << shift), htonl(upper_mask << shift));
    if (!high_bit)
      c->negate_expr_subtree(tree);
    high_bit_record = (high_bit_record << 1) | high_bit;
    count++;

    mask &= ~upper_mask;
    u &= mask;
  }

  while (count > 0) {
    c->finish_expr_subtree(tree, (high_bit_record & 1 ? Classifier::C_AND : Classifier::C_OR));
    high_bit_record >>= 1;
    count--;
  }

  if (_op_negated)
    c->negate_expr_subtree(tree);
}

void
IPFilter::Primitive::add_exprs(Classifier *c, Vector<int> &tree) const
{
  c->start_expr_subtree(tree);

  // enforce first fragment: fragmentation offset == 0
  // (before transport protocol to enhance later optimizations)
  if (_type == TYPE_PORT || _type == TYPE_TCPOPT || ((_type & TYPE_FIELD) && (_type & FIELD_PROTO_MASK)))
    c->add_expr(tree, 4, 0, htonl(0x00001FFF));
  
  // handle transport protocol uniformly
  if (_transp_proto != UNKNOWN)
    add_exprs_for_proto(_transp_proto, 0xFF, c, tree);

  // handle other types
  switch (_type) {

   case TYPE_HOST:
    c->start_expr_subtree(tree);
    if (_srcdst == SD_SRC || _srcdst == SD_AND || _srcdst == SD_OR)
      add_comparison_exprs(c, tree, 12, 0, true);
    if (_srcdst == SD_DST || _srcdst == SD_AND || _srcdst == SD_OR)
      add_comparison_exprs(c, tree, 16, 0, true);
    c->finish_expr_subtree(tree, (_srcdst == SD_OR ? C_OR : C_AND));
    break;

   case TYPE_PROTO:
    if (_transp_proto < 256)
      add_comparison_exprs(c, tree, 8, 16);
    break;

   case TYPE_IPFRAG:
    c->add_expr(tree, 4, 0, htonl(0x00003FFF));
    if (!_op_negated)
      c->negate_expr_subtree(tree);
    break;

   case TYPE_PORT:
    c->start_expr_subtree(tree);
    if (_srcdst == SD_SRC || _srcdst == SD_AND || _srcdst == SD_OR)
      add_comparison_exprs(c, tree, TRANSP_FAKE_OFFSET, 16);
    if (_srcdst == SD_DST || _srcdst == SD_AND || _srcdst == SD_OR)
      add_comparison_exprs(c, tree, TRANSP_FAKE_OFFSET, 0);
    c->finish_expr_subtree(tree, (_srcdst == SD_OR ? C_OR : C_AND));
    break;

   case TYPE_TCPOPT:
    c->add_expr(tree, TRANSP_FAKE_OFFSET + 12, htonl(_u.u << 16), htonl(_mask.u << 16));
    break;

   default:
    if (_type & TYPE_FIELD) {
      int offset = (_type & FIELD_OFFSET_MASK) >> FIELD_OFFSET_SHIFT;
      int length = ((_type & FIELD_LENGTH_MASK) >> FIELD_LENGTH_SHIFT) + 1;
      int word_offset = (offset >> 3) & ~3, bit_offset = offset & 0x1F;
      int transp_offset = (_type & FIELD_PROTO_MASK ? TRANSP_FAKE_OFFSET : 0);
      add_comparison_exprs(c, tree, transp_offset + word_offset, 32 - (bit_offset + length));
    } else
      assert(0);
    break;

  }

  c->finish_expr_subtree(tree);
}


static void
separate_text(const String &text, Vector<String> &words)
{
  const char* s = text.data();
  int len = text.length();
  int pos = 0;
  while (pos < len) {
    while (pos < len && isspace(s[pos]))
      pos++;
    switch (s[pos]) {

     case '&': case '|':
      if (pos < len - 1 && s[pos+1] == s[pos])
	goto two_char;
      goto one_char;

     case '<': case '>': case '!': case '=':
      if (pos < len - 1 && s[pos+1] == '=')
	goto two_char;
      goto one_char;
      
     case '(': case ')': case '[': case ']': case ',': case ';':
     case '?':
     one_char:
      words.push_back(text.substring(pos, 1));
      pos++;
      break;

     two_char:
      words.push_back(text.substring(pos, 2));
      pos += 2;
      break;

     default: {
	int first = pos;
	while (pos < len && (isalnum(s[pos]) || s[pos] == '-' || s[pos] == '.' || s[pos] == '/' || s[pos] == '@' || s[pos] == '_' || s[pos] == ':'))
	  pos++;
	if (pos == first)
	  pos++;
	words.push_back(text.substring(first, pos - first));
	break;
      }
      
    }
  }
}

void
IPFilter::notify_noutputs(int n)
{
  set_noutputs(n);
}

/*
 * expr ::= orexpr
 *	|   orexpr ? expr : expr
 *	;
 * orexpr ::= orexpr || orexpr
 *	|   orexpr or orexpr
 *	|   term
 *	;
 * term ::= term && term
 *	|   term and term
 *	|   term factor			// juxtaposition = and
 *	|   term
 * factor ::= ! factor
 *	|   true
 *	|   false
 *	|   quals data
 *	|   quals relop data
 *	|   ( expr )
 *	;
 */

int
IPFilter::parse_expr(const Vector<String> &words, int pos,
		     Vector<int> &tree, Primitive &prev_prim,
		     ErrorHandler *errh)
{
  start_expr_subtree(tree);

  while (1) {
    pos = parse_orexpr(words, pos, tree, prev_prim, errh);
    if (pos >= words.size())
      break;
    if (words[pos] != "?")
      break;
    int old_pos = pos + 1;
    pos = parse_expr(words, old_pos, tree, prev_prim, errh);
    if (pos > old_pos && pos < words.size() && words[pos] == ":")
      pos++;
    else {
      errh->error("':' missing in ternary expression");
      break;
    }
  }

  finish_expr_subtree(tree, C_TERNARY);
  return pos;
}

int
IPFilter::parse_orexpr(const Vector<String> &words, int pos,
		     Vector<int> &tree, Primitive &prev_prim,
		     ErrorHandler *errh)
{
  start_expr_subtree(tree);

  while (1) {
    pos = parse_term(words, pos, tree, prev_prim, errh);
    if (pos >= words.size())
      break;
    if (words[pos] == "or" || words[pos] == "||")
      pos++;
    else
      break;
  }

  finish_expr_subtree(tree, C_OR);
  return pos;
}

int
IPFilter::parse_term(const Vector<String> &words, int pos,
		     Vector<int> &tree, Primitive &prev_prim,
		     ErrorHandler *errh)
{
  start_expr_subtree(tree);

  bool blank_ok = false;
  while (1) {
    int next = parse_factor(words, pos, tree, prev_prim, false, errh);
    if (next == pos)
      break;
    blank_ok = true;
    if (next < words.size() && (words[next] == "and" || words[next] == "&&")) {
      blank_ok = false;
      next++;
    }
    pos = next;
  }

  if (!blank_ok)
    errh->error("missing term");
  finish_expr_subtree(tree);
  return pos;
}

static int
parse_brackets(IPFilter::Primitive& prim, const Vector<String>& words, int pos,
	       ErrorHandler* errh)
{
  int first_pos = pos + 1;
  String combination;
  for (pos++; pos < words.size() && words[pos] != "]"; pos++)
    combination += words[pos];
  if (pos >= words.size()) {
    errh->error("missing ']'");
    return first_pos;
  }
  pos++;

  // parse 'combination'
  int fieldpos, len = 1;
  const char* colon = find(combination.begin(), combination.end(), ':');
  const char* comma = find(combination.begin(), combination.end(), ',');
  if (colon < combination.end() - 1) {
    if (cp_integer(combination.begin(), colon, 0, &fieldpos) == colon
	&& cp_integer(colon + 1, combination.end(), 0, &len) == combination.end())
      goto non_syntax_error;
  } else if (comma < combination.end() - 1) {
    int pos2;
    if (cp_integer(combination.begin(), comma, 0, &fieldpos) == comma
	&& cp_integer(comma + 1, combination.end(), 0, &pos2) == combination.end()) {
      len = pos2 - fieldpos + 1;
      goto non_syntax_error;
    }
  } else if (cp_integer(combination, &fieldpos))
    goto non_syntax_error;
  errh->error("syntax error after '[', expected '[POS]' or '[POS:LEN]'");
  return pos;

 non_syntax_error:
  if (len < 1 || len > 4)
    errh->error("LEN in '[POS:LEN]' out of range, should be between 1 and 4");
  else if ((fieldpos & ~3) != ((fieldpos + len - 1) & ~3))
    errh->error("field [%d:%d] does not fit in a single word", fieldpos, len);
  else if (prim._transp_proto == IPFilter::UNKNOWN)
    prim.set_type(IPFilter::TYPE_FIELD | ((fieldpos*8) << IPFilter::FIELD_OFFSET_SHIFT) | ((len*8 - 1) << IPFilter::FIELD_LENGTH_SHIFT), errh);
  else
    prim.set_type(IPFilter::TYPE_FIELD | (prim._transp_proto << IPFilter::FIELD_PROTO_SHIFT) | ((fieldpos*8) << IPFilter::FIELD_OFFSET_SHIFT) | ((len*8 - 1) << IPFilter::FIELD_LENGTH_SHIFT), errh);
  return pos;
}

int
IPFilter::parse_factor(const Vector<String> &words, int pos,
		       Vector<int> &tree, Primitive &prev_prim,
		       bool negated, ErrorHandler *errh)
{
  // return immediately on last word, ")", "||", "or", "?", ":"
  if (pos >= words.size() || words[pos] == ")" || words[pos] == "||" || words[pos] == "or" || words[pos] == "?" || words[pos] == ":")
    return pos;

  // easy cases
  
  // 'true' and 'false'
  if (words[pos] == "true") {
    add_expr(tree, 0, 0, 0);
    if (negated)
      negate_expr_subtree(tree);
    return pos + 1;
  }
  if (words[pos] == "false") {
    add_expr(tree, 0, 0, 0);
    if (!negated)
      negate_expr_subtree(tree);
    return pos + 1;
  }
  // ! factor
  if (words[pos] == "not" || words[pos] == "!") {
    int next = parse_factor(words, pos + 1, tree, prev_prim, !negated, errh);
    if (next == pos + 1)
      errh->error("missing factor after '%s'", words[pos].c_str());
    return next;
  }
  // ( expr )
  if (words[pos] == "(") {
    int next = parse_expr(words, pos + 1, tree, prev_prim, errh);
    if (next == pos + 1)
      errh->error("missing expression after '('");
    if (next >= 0) {
      if (next >= words.size() || words[next] != ")")
	errh->error("missing ')'");
      else
	next++;
      if (negated)
	negate_expr_subtree(tree);
    }
    return next;
  }

  // hard case
  
  // expect quals [relop] data
  int first_pos = pos;
  Primitive prim;

  // collect qualifiers
  for (; pos < words.size(); pos++) {
    String wd = words[pos];
    uint32_t wdata;
    int wt = wordmap->lookup(wd, 0, UNKNOWN, wdata, 0);

    if (wt >= 0 && wt == TYPE_TYPE) {
      prim.set_type(wdata, errh);
      if ((wdata & TYPE_FIELD) && (wdata & FIELD_PROTO_MASK))
	prim.set_transp_proto((wdata & FIELD_PROTO_MASK) >> FIELD_PROTO_SHIFT, errh);

    } else if (wt >= 0 && wt == TYPE_PROTO)
      prim.set_transp_proto(wdata, errh);

    else if (wt != -1)
      break;

    else if (wd == "src") {
      if (pos < words.size() - 2 && (words[pos+2] == "dst" || words[pos+2] == "dest")) {
	if (words[pos+1] == "and" || words[pos+1] == "&&") {
	  prim.set_srcdst(SD_AND, errh);
	  pos += 2;
	} else if (words[pos+1] == "or" || words[pos+1] == "||") {
	  prim.set_srcdst(SD_OR, errh);
	  pos += 2;
	} else
	  prim.set_srcdst(SD_SRC, errh);
      } else
	prim.set_srcdst(SD_SRC, errh);
    } else if (wd == "dst" || wd == "dest")
      prim.set_srcdst(SD_DST, errh);
    
    else if (wd == "ip")
      /* nada */;
    
    else if (wd == "not" || wd == "!")
      negated = !negated;
      
    else
      break;
  }

  // prev_prim is not relevant if there were any qualifiers
  if (pos != first_pos)
    prev_prim.clear();

  // optional [] syntax
  String wd = (pos >= words.size() - 1 ? String() : words[pos]);
  if (wd == "[" && pos > first_pos && prim._type == TYPE_NONE) {
    pos = parse_brackets(prim, words, pos, errh);
    wd = (pos >= words.size() - 1 ? String() : words[pos]);
  }
  
  // optional relational operation
  pos++;
  if (wd == "=" || wd == "==")
    /* nada */;
  else if (wd == "!=")
    prim._op_negated = true;
  else if (wd == ">")
    prim._op = OP_GT;
  else if (wd == "<")
    prim._op = OP_LT;
  else if (wd == ">=") {
    prim._op = OP_LT;
    prim._op_negated = true;
  } else if (wd == "<=") {
    prim._op = OP_GT;
    prim._op_negated = true;
  } else
    pos--;
  
  // now collect the actual data
  if (pos < words.size()) {
    wd = words[pos];
    uint32_t wdata;
    int wt = wordmap->lookup(wd, prim._type, prim._transp_proto, wdata, errh);
    pos++;
    
    if (wt == -2)		// ambiguous or incorrect word type
      /* absorb word, but do nothing */
      prim._type = -2;

    else if (wt != -1 && wt != TYPE_TYPE) {
      prim._data = wt;
      prim._u.u = wdata;

    } else if (cp_integer(wd, &prim._u.i))
      prim._data = TYPE_INT;
  
    else if (cp_ip_address(wd, prim._u.c, this)) {
      if (pos < words.size() - 1 && words[pos] == "mask"
	  && cp_ip_address(words[pos+1], prim._mask.c, this)) {
	pos += 2;
	prim._data = TYPE_NET;
      } else if (prim._type == TYPE_NET && cp_ip_prefix(wd, prim._u.c, prim._mask.c, this))
	prim._data = TYPE_NET;
      else
	prim._data = TYPE_HOST;
    
    } else if (cp_ip_prefix(wd, prim._u.c, prim._mask.c, this))
      prim._data = TYPE_NET;

    else {
      if (prim._op != OP_EQ || prim._op_negated)
	errh->error("dangling operator near '%s'", wd.cc());
      pos--;
    }
  }

  if (pos == first_pos) {
    errh->error("empty term near '%s'", wd.cc());
    return pos;
  }
  
  // add if it is valid
  if (prim.check(prev_prim, errh) >= 0) {
    prim.add_exprs(this, tree);
    if (negated)
      negate_expr_subtree(tree);
    prev_prim = prim;
  }

  return pos;
}

int
IPFilter::configure(Vector<String> &conf, ErrorHandler *errh)
{
  int before_nerrors = errh->nerrors();
  _output_everything = -1;

  // requires packet headers be aligned
  _align_offset = 0;

  Vector<int> tree;
  init_expr_subtree(tree);
  
  // [QUALS] [host|net|port|proto] [data]
  // QUALS ::= src | dst | src and dst | src or dst | \empty
  //        |  ip | icmp | tcp | udp
  for (int argno = 0; argno < conf.size(); argno++) {
    Vector<String> words;
    separate_text(cp_unquote(conf[argno]), words);

    if (words.size() == 0) {
      errh->error("empty pattern %d", argno);
      continue;
    }

    PrefixErrorHandler cerrh(errh, "pattern " + String(argno) + ": ");
    
    // get slot
    int slot = noutputs();
    {
      String slotwd = words[0];
      if (slotwd == "allow") {
	slot = 0;
	if (noutputs() == 0)
	  cerrh.error("'allow' is meaningless, element has zero outputs");
      } else if (slotwd == "deny") {
	slot = noutputs();
	if (noutputs() > 1)
	  cerrh.warning("meaning of 'deny' has changed (now it means 'drop')");
      } else if (slotwd == "drop")
	slot = noutputs();
      else if (cp_integer(slotwd, &slot)) {
	if (slot < 0 || slot >= noutputs()) {
	  cerrh.error("slot '%d' out of range", slot);
	  slot = noutputs();
	}
      } else
	cerrh.error("unknown slot ID '%s'", slotwd.cc());
    }

    start_expr_subtree(tree);

    // check for "-"
    if (words.size() == 1 || (words.size() == 2 && words[1] == "-")
	|| (words.size() == 2 && words[1] == "any")
	|| (words.size() == 2 && words[1] == "all"))
      add_expr(tree, 0, 0, 0);

    else {
      // start with a blank primitive
      Primitive prev_prim;
      
      int pos = parse_expr(words, 1, tree, prev_prim, &cerrh);
      if (pos < words.size())
	cerrh.error("garbage after expression at '%s'", words[pos].c_str());
    }
    
    finish_expr_subtree(tree, C_AND, -slot);
  }

  finish_expr_subtree(tree, C_OR, -noutputs(), -noutputs());
  
  //{ String sxxx = program_string(this, 0); click_chatter("%s", sxxx.cc()); }
  optimize_exprs(errh);
  //{ String sxxx = program_string(this, 0); click_chatter("%s", sxxx.cc()); }
  return (errh->nerrors() == before_nerrors ? 0 : -1);
}


//
// RUNNING
//

void
IPFilter::length_checked_push(Packet *p)
{
  const unsigned char *neth_data = (const unsigned char *)p->ip_header();
  const unsigned char *transph_data = (const unsigned char *)p->transport_header();
  int packet_length = p->length() + TRANSP_FAKE_OFFSET - p->transport_header_offset();
  Expr *ex = &_exprs[0];	// avoid bounds checking
  int pos = 0;
  
  do {
    int off = ex[pos].offset;
    unsigned data;
    if (off + 4 > packet_length)
      goto check_length;
    
   length_ok:
    if (off >= TRANSP_FAKE_OFFSET)
      data = *(const unsigned *)(transph_data + off - TRANSP_FAKE_OFFSET);
    else
      data = *(const unsigned *)(neth_data + off);
    if ((data & ex[pos].mask.u)	== ex[pos].value.u)
      pos = ex[pos].yes;
    else
      pos = ex[pos].no;
    continue;
    
   check_length:
    if (ex[pos].offset < packet_length) {
      unsigned available = packet_length - ex[pos].offset;
      if (!(ex[pos].mask.c[3]
	    || (ex[pos].mask.c[2] && available <= 2)
	    || (ex[pos].mask.c[1] && available == 1)))
	goto length_ok;
    }
    pos = ex[pos].no;
  } while (pos > 0);
  
  checked_output_push(-pos, p);
}

void
IPFilter::push(int, Packet *p)
{
  const unsigned char *neth_data = (const unsigned char *)p->ip_header();
  const unsigned char *transph_data = (const unsigned char *)p->transport_header();
  Expr *ex;
  int pos = 0;
  
  if (_output_everything >= 0) {
    // must use checked_output_push because the output number might be
    // out of range
    pos = -_output_everything;
    goto found;
  } else if (p->length() + TRANSP_FAKE_OFFSET - p->transport_header_offset() < _safe_length) {
    // common case never checks packet length
    length_checked_push(p);
    return;
  }
  
  ex = &_exprs[0];	// avoid bounds checking
  
  do {
    int off = ex[pos].offset;
    unsigned data;
    if (off >= TRANSP_FAKE_OFFSET)
      data = *(const unsigned *)(transph_data + off - TRANSP_FAKE_OFFSET);
    else
      data = *(const unsigned *)(neth_data + off);
    if ((data & ex[pos].mask.u)	== ex[pos].value.u)
      pos = ex[pos].yes;
    else
      pos = ex[pos].no;
  } while (pos > 0);
  
 found:
  checked_output_push(-pos, p);
}

CLICK_ENDDECLS
ELEMENT_REQUIRES(Classifier)
EXPORT_ELEMENT(IPFilter)
